// Description: 
// By Sarah Dean
// Email: sdean12@sdean12.org
// WWW:   http://www.FreeOTFE.org/
//
// -----------------------------------------------------------------------------
//


#include "FreeOTFECypherImpl.h"
#include "FreeOTFECypherXOR.h"

#include "FreeOTFEDebug.h"
#include "FreeOTFElib.h"


// =========================================================================
// Cypher driver init function
// devExt - The device extension to be initialized
NTSTATUS
ImpCypherDriverExtDetailsInit(
    IN OUT DEVICE_EXTENSION* devExt
)
{
    NTSTATUS status = STATUS_SUCCESS;
    int idx = -1;

    DEBUGOUTCYPHERIMPL(DEBUGLEV_ENTER, ("ImpCypherDriverExtDetailsInit\n"));


    // -- POPULATE DRIVER IDENTIFICATION --
    RtlZeroMemory(devExt->DriverTitle, sizeof(devExt->DriverTitle));
    RtlCopyMemory(
                  devExt->DriverTitle,
                  DRIVER_TITLE,
                  strlen(DRIVER_TITLE)
                 );

    devExt->DriverGUID = DRIVER_GUID;
    devExt->DriverVersionID = DRIVER_CYPHER_VERSION;


    // -- POPULATE cyphers SUPPORTED --
    devExt->CypherDetails = ExAllocatePool(
                                         NonPagedPool,
                                         (sizeof(CYPHER) * CYPHERS_SUPPORTED)
                                        );    

    devExt->CypherCount = CYPHERS_SUPPORTED;



    // -- XOR --
    idx++;
    RtlZeroMemory(devExt->CypherDetails[idx].Title, MAX_CYPHER_TITLE);
    RtlCopyMemory(
                  devExt->CypherDetails[idx].Title,
                  DRIVER_CIPHER_TITLE_XOR,
                  strlen(DRIVER_CIPHER_TITLE_XOR)
                 );

    devExt->CypherDetails[idx].Mode       = CYPHER_MODE_NONE;
    devExt->CypherDetails[idx].KeySize    = -1;
    devExt->CypherDetails[idx].BlockSize  = -1;
    devExt->CypherDetails[idx].VersionID  = DRIVER_CYPHER_IMPL_VERSION;
    devExt->CypherDetails[idx].CypherGUID = CIPHER_GUID_XOR;



    DEBUGOUTCYPHERIMPL(DEBUGLEV_EXIT, ("ImpCypherDriverExtDetailsInit\n"));

    return status;
}


// =========================================================================
// Cypher driver cleardown function
// devExt - The device extension to be cleared down
NTSTATUS
ImpCypherDriverExtDetailsCleardown(    
    IN OUT DEVICE_EXTENSION* devExt
)
{
    NTSTATUS status = STATUS_SUCCESS;

    DEBUGOUTCYPHERIMPL(DEBUGLEV_ENTER, ("ImpCypherDriverExtDetailsCleardown\n"));

    if (devExt->CypherDetails != NULL)
        {
        ExFreePool(devExt->CypherDetails);
        }

    devExt->CypherDetails = NULL;
    devExt->CypherCount = 0;

    DEBUGOUTCYPHERIMPL(DEBUGLEV_EXIT, ("ImpCypherDriverExtDetailsCleardown\n"));

    return status;
}


// =========================================================================
// Encryption function
// Note: CyphertextLength must be set to the size of the CyphertextData buffer on
//       entry; on exit, this will be set to the size of the buffer used.
NTSTATUS
ImpCypherEncryptData(
    IN      GUID* CypherGUID,
    IN      int KeyLength,  // In bits
    IN      char *Key,
    IN      char *KeyASCII,  // ASCII representation of "Key"
    IN      int IVLength,  // In bits
    IN      char *IV,
    IN      int PlaintextLength,  // In bytes
    IN      char *PlaintextData,
    OUT     char *CyphertextData
)
{
    NTSTATUS status = STATUS_SUCCESS;

    DEBUGOUTCYPHERIMPL(DEBUGLEV_ENTER, ("ImpCypherEncryptData\n"));

    if (!(IsEqualGUID(&CIPHER_GUID_XOR, CypherGUID)))
        {
        status = STATUS_INVALID_PARAMETER;
        }
    else
        {
        if ( (IV == NULL) && (IVLength != 0) )
            {
            // Sanity check failed...
            DEBUGOUTCYPHERIMPL(DEBUGLEV_INFO, ("Inconsistant: NULL passed as %d bit long IV\n", IVLength));
            status = STATUS_INVALID_PARAMETER;
            }
        else
            {
            _xor_data(KeyLength, Key, PlaintextLength, PlaintextData, CyphertextData);
            }
        }

    DEBUGOUTCYPHERIMPL(DEBUGLEV_EXIT, ("ImpCypherEncryptData\n"));

    return status;
}


// =========================================================================
// Decryption function
// Note: PlaintextLength must be set to the size of the PlaintextData buffer on
//       entry; on exit, this will be set to the size of the buffer used.
NTSTATUS
ImpCypherDecryptData(
    IN      GUID* CypherGUID,
    IN      int KeyLength,  // In bits
    IN      char *Key,
    IN      char *KeyASCII,  // ASCII representation of "Key"
    IN      int IVLength,  // In bits
    IN      char *IV,
    IN      int CyphertextLength,  // In bytes
    IN      char *CyphertextData,
    OUT     char *PlaintextData
)
{
    NTSTATUS status = STATUS_SUCCESS;

    DEBUGOUTCYPHERIMPL(DEBUGLEV_ENTER, ("ImpCypherDecryptData\n"));
    DEBUGOUTCYPHERIMPL(DEBUGLEV_INFO, ("key length = %d; data length = %d\n", KeyLength, CyphertextLength));

    if (!(IsEqualGUID(&CIPHER_GUID_XOR, CypherGUID)))
        {
        status = STATUS_INVALID_PARAMETER;
        }
    else
        {
        if ( (IV == NULL) && (IVLength != 0) )
            {
            // Sanity check failed...
            DEBUGOUTCYPHERIMPL(DEBUGLEV_INFO, ("Inconsistant: NULL passed as %d bit long IV\n", IVLength));
            status = STATUS_INVALID_PARAMETER;
            }
        else
            {
            _xor_data(KeyLength, Key, CyphertextLength, CyphertextData, PlaintextData);
            }
        }

    DEBUGOUTCYPHERIMPL(DEBUGLEV_EXIT, ("ImpCypherDecryptData\n"));

    return status;
}


// =========================================================================
// Internal function to carry out actual XOR encryption
NTSTATUS
_xor_data(
    IN      int KeyLength,  // In bits
    IN      char *Key,
    IN      int DataLength,  // In bytes
    IN      char *InData,
    OUT     char *OutData
)
{
    NTSTATUS status = STATUS_SUCCESS;
    int i;
    int keyOffset;
    int keyLenBytes;

    DEBUGOUTCYPHERIMPL(DEBUGLEV_ENTER, ("_xor_data\n"));

    keyLenBytes = (KeyLength / 8);
    // Sanity check, just in case we were passed an empty password!
    if (keyLenBytes != 0)
        {
        keyOffset = 0;
        for(i = 0; i < DataLength; i++, keyOffset++)
            {
            if (keyOffset >= keyLenBytes)
                {
                keyOffset = 0;
                }

            OutData[i] = InData[i] ^ Key[keyOffset];
            }
        }
    else
        {
        // No password? Just copy the data from the input buffer to the output
        RtlCopyMemory(
                      OutData,
                      InData,
                      DataLength
                     );
        }

    DEBUGOUTCYPHERIMPL(DEBUGLEV_EXIT, ("_xor_data\n"));

    return status;
}


// =========================================================================
// =========================================================================

